// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_CRVSRF_COORDFUNCTIONIMPLEMENTATIONS_HH
#define DUNE_CRVSRF_COORDFUNCTIONIMPLEMENTATIONS_HH

#include <dune/curvedsurfacegrid/surfacedistance/surfacedistance.hh>
#include <dune/curvedsurfacegrid/surfacedistance/vertexmap.hh>
#include <dune/curvedsurfacegrid/surfacedistance/vtureader.hh>
#include <dune/grid/geometrygrid/coordfunction.hh>

#define COLLECTOR_BENCHMARK 0

#if COLLECTOR_BENCHMARK != 0
#include <ctime>
//wallclock-time in seconds
double dbl_time(){
  #ifdef NO_CLOCK_GETTIME
    return (double)time(NULL);
  #else
    struct timespec now;
    clock_gettime(CLOCK_MONOTONIC, &now);
    return (double)now.tv_nsec / 1000000000.0 + now.tv_sec;
  #endif
}
#endif

namespace Dune
{

  //CenterSphereCoordFunction
  struct CenterSphereCoordFunction
  : public AnalyticalCoordFunction<double, 3, 3, CenterSphereCoordFunction>
  {
    //constrcutor
    CenterSphereCoordFunction(double radius) : radius_(radius){}

    //operator()
    FieldVector<double, 3> operator()(FieldVector<double, 3> x) const{
      #if COLLECTOR_BENCHMARK != 0
        inputs.push_back(x);
      #endif
      x *= radius_ / x.two_norm();
      return x;
    }

    #if COLLECTOR_BENCHMARK != 0
    void collectorBenchmark(bool fresh_cache){
      double start_time = dbl_time();
      for(auto &input : inputs){
        input *= radius_ / input.two_norm();
      }
      double run_time = dbl_time() - start_time;
      std::cout << "csphere time: " << crvsrf::ft(run_time, 11, 9)
                      << ", count: " << inputs.size() << std::endl;
    }
    #endif

  private:
    double radius_;
    #if COLLECTOR_BENCHMARK != 0
      mutable std::vector<FieldVector<double, 3> > inputs;
    #endif
  };



  //SphereCoordFunction
  struct SphereCoordFunction
  : public AnalyticalCoordFunction<double, 3, 3, SphereCoordFunction>
  {
    //constrcutor
    SphereCoordFunction(FieldVector<double, 3> center, double radius)
      : center_(center), radius_(radius){}

    //operator()
    FieldVector<double, 3> operator()(FieldVector<double, 3> x) const{
      #if COLLECTOR_BENCHMARK != 0
        inputs.push_back(x);
      #endif
      FieldVector<double, 3> direction = x - center_;
      double factor = radius_ / direction.two_norm();
      for(int i = 0; i < 3; ++i) x[i] = center_[i] + direction[i] * factor;
      return x;
    }

    #if COLLECTOR_BENCHMARK != 0
    void collectorBenchmark(bool fresh_cache){
      double start_time = dbl_time();
      for(auto &input : inputs){
        FieldVector<double, 3> direction = input - center_;
        double factor = radius_ / direction.two_norm();
        for(int i = 0; i < 3; ++i) input[i] = center_[i] + direction[i] * factor;
      }
      double run_time = dbl_time() - start_time;
      std::cout << "sphere time: " << crvsrf::ft(run_time, 11, 9)
                     << ", count: " << inputs.size() << std::endl;
    }
    #endif

  private:
    FieldVector<double, 3> center_;
    double radius_;
    #if COLLECTOR_BENCHMARK != 0
      mutable std::vector<FieldVector<double, 3> > inputs;
    #endif
  };



  //SurfaceDistanceCoordFunction
  template<bool cached=true>
  struct SurfaceDistanceCoordFunction
  : public AnalyticalCoordFunction<double, 3, 3, SurfaceDistanceCoordFunction<cached> >
  {
    const bool coordCaching = cached;

    //constructor using vtu-reader to read "file"
    SurfaceDistanceCoordFunction(const std::string &file, double expansion_width=-1.0,
                                                bool no_scaling=false) : cache(nullptr){
      surface_mesh = new crvsrf::Mesh(2, 3);
      read_vtu_file(file, *surface_mesh);
      backgrid = new crvsrf::BackGrid(*surface_mesh, expansion_width, no_scaling);
      crvsrf::Vertex c_begin, c_end;
      surface_mesh->determine_coordinate_extrema(c_begin, c_end);
      crvsrf::Vertex c_expansion = (c_end - c_begin) * 0.1;
      if(cached){
        cache = new crvsrf::VertexMap<crvsrf::Vertex>(c_begin - c_expansion, c_end + c_expansion, 3);
      }
    }

    //constructor using a pre-loaded Dune-surface-mesh
    /*!SurfaceDistanceCoordFunction(Grid &mesh, double expansion_width=-1.0, bool no_scaling=false){
      surface_mesh = convert_dune_mesh_to_crvsrf::mesh...
      backgrid = new BackGrid(*surface_mesh, expansion_width, no_scaling);
      //! implement cache-init
    }*/

    //destructor
    ~SurfaceDistanceCoordFunction(){
      if(cached) delete cache;
      delete backgrid;
      delete surface_mesh;
    }

    //resetCache
    void resetCache(crvsrf::Vertex begin = crvsrf::Vertex(),
                     crvsrf::Vertex end = crvsrf::Vertex(),
                     double expansion_factor = 0.1){
      if(cached){
        delete cache;
        if(begin.is_origin() && end.is_origin()){
          surface_mesh->determine_coordinate_extrema(begin, end);
        }
        crvsrf::Vertex c_expansion = (end - begin) * expansion_factor;
        cache = new crvsrf::VertexMap<crvsrf::Vertex>(begin - c_expansion, end + c_expansion, 3);
      }else{
        crvsrf::error("Cannot reset cache in uncached SurfaceDistanceCoordFunction!");
      }
    }

    //operator()
    FieldVector<double, 3> operator()(const FieldVector<double, 3> &x) const;

    //output_backgrid_dimensions
    void output_backgrid_dimensions(){
      backgrid->debug_output_dimensions();
    }

    #if COLLECTOR_BENCHMARK != 0
    void collectorBenchmark(bool fresh_cache){
      if(cached){
        if(fresh_cache) cache->clear();
        int uncached_cnt = 0;
        double start_time = dbl_time();
        for(auto &input : inputs){
          crvsrf::Vertex hit_vertex;
          crvsrf::Vertex v(input[0], input[1], input[2]);
          if(!cache->find_3d(v, hit_vertex)){
            backgrid->distance_to_surface_3d(v, hit_vertex);
            cache->insert(v, hit_vertex);
            ++uncached_cnt;
          }
        }
        double run_time = dbl_time() - start_time;
        std::cout << "surfdist time: " << crvsrf::ft(run_time, 11, 9) << ", count: "<< inputs.size()
                                                     << ", uncached: " << uncached_cnt << std::endl;
      }else{ //uncached
        double start_time = dbl_time();
        for(auto &input : inputs){
          crvsrf::Vertex hit_vertex;
          crvsrf::Vertex v(input[0], input[1], input[2]);
          backgrid->distance_to_surface_3d(v, hit_vertex);
        }
        double run_time = dbl_time() - start_time;
        std::cout << "surfdist time: " << crvsrf::ft(run_time, 11, 9)
                          << ", count: "<< inputs.size() << std::endl;
      }
    }
    #endif

    #if VERTEXMAP_DEBUG_MODE != 0
    void output_vertexmap_statistics(){
      if(cached){
        cache->debug_output_statistics();
      }else{
        crvsrf::error("Cannot output vertexmap-statistics in "
                      "uncached SurfaceDistanceCoordFunction!");
      }
    }
    #endif

  private:
    crvsrf::Mesh *surface_mesh;
    crvsrf::BackGrid *backgrid;
    crvsrf::VertexMap<crvsrf::Vertex> *cache;
    #if COLLECTOR_BENCHMARK != 0
      mutable std::vector<FieldVector<double, 3> > inputs;
    #endif
  };


  //operator() (cached)
  template<>
  inline FieldVector<double, 3> SurfaceDistanceCoordFunction<true>::operator()(
                                                const FieldVector<double, 3> &x) const{
    #if COLLECTOR_BENCHMARK != 0
      inputs.push_back(x);
    #endif
    crvsrf::Vertex hit_vertex;
    crvsrf::Vertex v(x[0], x[1], x[2]);
    if(!cache->find_3d(v, hit_vertex)){
      backgrid->distance_to_surface_3d(v, hit_vertex);
      cache->insert(v, hit_vertex);
    }
    return {hit_vertex[0], hit_vertex[1], hit_vertex[2]};
  }


  //operator() (uncached)
  template<>
  inline FieldVector<double, 3> SurfaceDistanceCoordFunction<false>::operator()(
                                                 const FieldVector<double, 3> &x) const{
    #if COLLECTOR_BENCHMARK != 0
      inputs.push_back(x);
    #endif
    crvsrf::Vertex hit_vertex;
    crvsrf::Vertex v(x[0], x[1], x[2]);
    backgrid->distance_to_surface_3d(v, hit_vertex);
    return {hit_vertex[0], hit_vertex[1], hit_vertex[2]};
  }

} // namespace Dune

#endif // #ifndef DUNE_CRVSRF_COORDFUNCTIONIMPLEMENTATIONS_HH
