// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_CRVSRF_COORDPROVIDER_HH
#define DUNE_CRVSRF_COORDPROVIDER_HH

#include <array>

namespace Dune
{

  namespace crvsrf
  {

    template< class Coordinate>
    Coordinate edgeCenter(Coordinate gc1, const Coordinate &gc2)
    {
      for(typename Coordinate::size_type i = 0, end = gc1.size(); i < end; ++i)
      {
        gc1[i] = (gc1[i] + gc2[i]) / 2.0;
      }
      return gc1;
    }

    template< class Coordinate>
    Coordinate edgeVertex(Coordinate gc1, const Coordinate &gc2, double factor)
    {
      for(typename Coordinate::size_type i = 0, end = gc1.size(); i < end; ++i)
      {
        gc1[i] = gc1[i] + (gc2[i] - gc1[i]) * factor;
      }
      return gc1;
    }

    template< class Coordinate>
    Coordinate faceVertex(Coordinate gc1, const Coordinate &gc2, const Coordinate &gc3,
                                            double coeff1, double coeff2, double coeff3)
    {
      for(typename Coordinate::size_type i = 0, end = gc1.size(); i < end; ++i)
      {
        gc1[i] = coeff1 * gc1[i] + coeff2 * gc2[i] + coeff3 * gc3[i];
      }
      return gc1;
    }



    // InterpolatoryVerticesGenerator
    // ------------------------------

    template< class Coordinate, int mydim, int order >
    struct InterpolatoryVerticesGenerator;

    //Vertex
    template< class Coordinate, int order >
    struct InterpolatoryVerticesGenerator< Coordinate, 0, order >
    {
      static void generate(const std::array<Coordinate, 1> &corners, std::vector<Coordinate> &vertices)
      {
        vertices.resize(1);
        vertices[0] = corners[0];
      }
    };

    //Edge
    template< class Coordinate, int order >
    struct InterpolatoryVerticesGenerator< Coordinate, 1, order >
    {
      static void generate(const std::array<Coordinate, 2> &corners, std::vector<Coordinate> &vertices)
      {
        vertices.resize(order + 1);
        vertices[0] = corners[0];
        for(int i = 1; i < order; ++i)
        {
          vertices[i] = edgeVertex(corners[0], corners[1], (double)i / order);
        }
        vertices[order] = corners[1];
      }
    };

    //Triangle (order 1)
    template< class Coordinate >
    struct InterpolatoryVerticesGenerator< Coordinate, 2, 1 >
    {
      static void generate(const std::array<Coordinate, 4> &corners, std::vector<Coordinate> &vertices)
      {
        vertices.resize(3);
        vertices[0] = corners[0];
        vertices[1] = corners[1];
        vertices[2] = corners[2];
      }
    };

    //Triangle (order 2)
    template< class Coordinate >
    struct InterpolatoryVerticesGenerator< Coordinate, 2, 2 >
    {
      static void generate(const std::array<Coordinate, 4> &corners, std::vector<Coordinate> &vertices)
      {
        vertices.resize(6);
        vertices[0] = corners[0];
        vertices[1] = edgeCenter(corners[0], corners[1]);
        vertices[2] = corners[1];
        vertices[3] = edgeCenter(corners[0], corners[2]);
        vertices[4] = edgeCenter(corners[1], corners[2]);
        vertices[5] = corners[2];
      }
    };

    //Triangle (order 3)
    template< class Coordinate >
    struct InterpolatoryVerticesGenerator< Coordinate, 2, 3 >
    {
      static void generate(const std::array<Coordinate, 4> &corners, std::vector<Coordinate> &vertices)
      {
        static const double cf1 = 1.0 / 3.0;
        static const double cf2 = 2.0 / 3.0;
        vertices.resize(10);
        vertices[0] = corners[0];
        vertices[1] = edgeVertex(corners[0], corners[1], cf1);
        vertices[2] = edgeVertex(corners[0], corners[1], cf2);
        vertices[3] = corners[1];
        vertices[4] = edgeVertex(corners[0], corners[2], cf1);
        vertices[5] = faceVertex(corners[0], corners[1], corners[2], cf1, cf1, cf1);
        vertices[6] = edgeVertex(corners[1], corners[2], cf1);
        vertices[7] = edgeVertex(corners[0], corners[2], cf2);
        vertices[8] = edgeVertex(corners[1], corners[2], cf2);
        vertices[9] = corners[2];
      }
    };

    //Triangle (order 4)
    template< class Coordinate >
    struct InterpolatoryVerticesGenerator< Coordinate, 2, 4 >
    {
      static void generate(const std::array<Coordinate, 4> &corners, std::vector<Coordinate> &vertices)
      {
        static const double cf1 = 1.0 / 4.0;
        static const double cf2 = 2.0 / 4.0;
        static const double cf3 = 3.0 / 4.0;
        vertices.resize(15);
        vertices[0] = corners[0];
        vertices[1] = edgeVertex(corners[0], corners[1], cf1);
        vertices[2] = edgeCenter(corners[0], corners[1]);
        vertices[3] = edgeVertex(corners[0], corners[1], cf3);
        vertices[4] = corners[1];
        vertices[5] = edgeVertex(corners[0], corners[2], cf1);
        vertices[6] = faceVertex(corners[0], corners[1], corners[2], cf2, cf1, cf1);
        vertices[7] = faceVertex(corners[0], corners[1], corners[2], cf1, cf2, cf1);
        vertices[8] = edgeVertex(corners[1], corners[2], cf1);
        vertices[9] = edgeCenter(corners[0], corners[2]);
        vertices[10] = faceVertex(corners[0], corners[1], corners[2], cf1, cf1, cf2);
        vertices[11] = edgeCenter(corners[1], corners[2]);
        vertices[12] = edgeVertex(corners[0], corners[2], cf3);
        vertices[13] = edgeVertex(corners[1], corners[2], cf3);
        vertices[14] = corners[2];
      }
    };

    //Triangle (order 5)
    template< class Coordinate >
    struct InterpolatoryVerticesGenerator< Coordinate, 2, 5 >
    {
      static void generate(const std::array<Coordinate, 4> &corners, std::vector<Coordinate> &vertices)
      {
        static const double cf1 = 1.0 / 5.0;
        static const double cf2 = 2.0 / 5.0;
        static const double cf3 = 3.0 / 5.0;
        static const double cf4 = 4.0 / 5.0;
        vertices.resize(21);
        vertices[0] = corners[0];
        vertices[1] = edgeVertex(corners[0], corners[1], cf1);
        vertices[2] = edgeVertex(corners[0], corners[1], cf2);
        vertices[3] = edgeVertex(corners[0], corners[1], cf3);
        vertices[4] = edgeVertex(corners[0], corners[1], cf4);
        vertices[5] = corners[1];
        vertices[6] = edgeVertex(corners[0], corners[2], cf1);
        vertices[7] = faceVertex(corners[0], corners[1], corners[2], cf3, cf1, cf1);
        vertices[8] = faceVertex(corners[0], corners[1], corners[2], cf2, cf2, cf1);
        vertices[9] = faceVertex(corners[0], corners[1], corners[2], cf1, cf3, cf1);
        vertices[10] = edgeVertex(corners[1], corners[2], cf1);
        vertices[11] = edgeVertex(corners[0], corners[2], cf2);
        vertices[12] = faceVertex(corners[0], corners[1], corners[2], cf2, cf1, cf2);
        vertices[13] = faceVertex(corners[0], corners[1], corners[2], cf1, cf2, cf2);
        vertices[14] = edgeVertex(corners[1], corners[2], cf2);
        vertices[15] = edgeVertex(corners[0], corners[2], cf3);
        vertices[16] = faceVertex(corners[0], corners[1], corners[2], cf1, cf1, cf3);
        vertices[17] = edgeVertex(corners[1], corners[2], cf3);
        vertices[18] = edgeVertex(corners[0], corners[2], cf4);
        vertices[19] = edgeVertex(corners[1], corners[2], cf4);
        vertices[20] = corners[2];
      }
    };



    // CoordProvider (adapted CoordVector from GeometryGrid)
    // -----------------------------------------------------

    template< int mydim, class Grid, bool fake >
    class CoordProvider;


    //CoordProvider (real)
    template< int mydim, class Grid >
    class CoordProvider< mydim, Grid, false >
    {
      typedef typename std::remove_const< Grid >::type::Traits Traits;

      typedef typename Traits::ctype ctype;

      static const int dimension = Traits::dimension;
      static const int mydimension = mydim;
      static const int codimension = dimension - mydimension;
      static const int dimensionworld = Traits::dimensionworld;

      typedef FieldVector< ctype, dimensionworld > Coordinate;

      typedef typename Traits::HostGrid HostGrid;
      typedef typename Traits::CoordFunction CoordFunction;

      typedef typename HostGrid::template Codim< codimension >::Entity HostEntity;

    public:
      CoordProvider ( const HostEntity &hostEntity,
                    const CoordFunction &coordFunction )
        : coordFunction_(coordFunction),
          hostEntity_(hostEntity)
      {}

      void calculate ( std::vector< Coordinate > &vertices ) const
      {
        const std::size_t numCorners = hostEntity_.geometry().corners();
        std::array<Coordinate, (1 << mydim)> corners;
        for( std::size_t i = 0; i < numCorners; ++i )
          corners[i] = hostEntity_.geometry().corner(i);
        InterpolatoryVerticesGenerator<Coordinate, mydim, Grid::order>::generate(corners, vertices);
        for(auto &v : vertices) coordFunction_.evaluate(v, v);
      }

    private:
      const CoordFunction &coordFunction_;
      const HostEntity &hostEntity_;
    };

    //CoordProvider (fake)
    template< int mydim, class Grid >
    class CoordProvider< mydim, Grid, true >
    {
      typedef typename std::remove_const< Grid > :: type :: Traits Traits;

      typedef typename Traits::ctype ctype;

      static const int dimension = Traits::dimension;
      static const int mydimension = mydim;
      static const int codimension = dimension - mydimension;
      static const int dimensionworld = Traits::dimensionworld;

      typedef FieldVector< ctype, dimensionworld > Coordinate;

      typedef typename Traits::HostGrid HostGrid;
      typedef typename Traits::CoordFunction CoordFunction;

      typedef typename HostGrid::template Codim< 0 >::Entity HostElement;

    public:
      CoordProvider ( const HostElement &hostElement,
                    const unsigned int subEntity,
                    const CoordFunction &coordFunction )
        : coordFunction_(coordFunction),
          hostElement_(hostElement),
          subEntity_( subEntity )
      {}

      void calculate ( std::vector< Coordinate > &vertices ) const
      {
        const GeometryType type = hostElement_.geometry().type();
        auto refElement = referenceElement< ctype, dimension >( type );
        const std::size_t numCorners = refElement.size( subEntity_, codimension, dimension );
        std::array<Coordinate, (1 << mydim)> corners;
        for( std::size_t i = 0; i < numCorners; ++i )
        {
          const std::size_t j = refElement.subEntity( subEntity_, codimension, i, dimension );
          corners[i] = hostElement_.geometry().corner(j);
        }
        InterpolatoryVerticesGenerator<Coordinate, mydim, Grid::order>::generate(corners, vertices);
        for(auto &v : vertices) coordFunction_.evaluate(v, v);
      }

    private:
      const CoordFunction &coordFunction_;
      const HostElement &hostElement_;
      const unsigned int subEntity_;
    };



    // IntersectionCoordProvider
    // -----------------------

    template< class Grid >
    class IntersectionCoordProvider
    {
      typedef typename std::remove_const< Grid >::type::Traits Traits;

      typedef typename Traits::ctype ctype;

      static const int dimension = Traits::dimension;
      static const int codimension = 1;
      static const int mydimension = dimension-codimension;
      static const int dimensionworld = Traits::dimensionworld;

      typedef FieldVector< ctype, dimensionworld > Coordinate;

      typedef typename Traits::HostGrid HostGrid;
      typedef typename Traits::CoordFunction CoordFunction;

      typedef typename Traits::template Codim< 0 >::GeometryImpl ElementGeometryImpl;
      typedef typename Traits::template Codim< codimension >::LocalGeometry HostLocalGeometry;

    public:
      IntersectionCoordProvider ( const ElementGeometryImpl &elementGeometry,
                                  const HostLocalGeometry &hostLocalGeometry,
                                  const CoordFunction &coordFunction )
        : coordFunction_(coordFunction),
          elementGeometry_( elementGeometry ),
          hostLocalGeometry_( hostLocalGeometry )
      {}

      void calculate ( std::vector< Coordinate > &vertices ) const
      {
        const std::size_t numCorners = hostLocalGeometry_.corners();
        std::array<Coordinate, (1 << mydimension)> corners;
        // $flo: coordFunction is applied here which is in contrast to GeomentryGrid's behaviour!
        for( std::size_t i = 0; i < numCorners; ++i )
          corners[ i ] = elementGeometry_.global( hostLocalGeometry_.corner( i ) );
        InterpolatoryVerticesGenerator<Coordinate, mydimension, Grid::order>::generate(corners, vertices);
        for(auto &v : vertices) coordFunction_.evaluate(v, v);
      }

      template< unsigned int numCorners >
      void calculate ( Coordinate (&corners)[ numCorners ] ) const
      {
        assert( numCorners == hostLocalGeometry_.corners() );
      }

    private:
      const CoordFunction &coordFunction_;
      const ElementGeometryImpl &elementGeometry_;
      HostLocalGeometry hostLocalGeometry_;
    };

  } // namespace crvsrf

} // namespace Dune

#endif // #ifndef DUNE_CRVSRF_COORDPROVIDER_HH
